'' =================================================================================================
''
''   File....... jm_hc-8+_vixen_player.spin
''   Purpose.... Play 8-channel VSA show; live from PC or from microSD card
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2012 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 17 JUL 2012
''
'' =================================================================================================

{{

  This is a dual-purpose program:

  1) Play servo and dimmer data streamed from Vixen (@115_200 baud)
  2) Play binary file exported from VSA
     -- requires EFX-TEK microSD adapter (#99002) and microSD card to hold files

  This program limits the total channel count to 8 (for single HC-8+). The program supports
  a mix of servos and LED dimmer channels, again the total count is eight. If servos are used
  they start on channel 0 and can go up to channel 7 (assuming no LED dimmers used).

  Possible configurations:

    Servos/Channels    Dimmers/Channels
    ---------------    ----------------
       0 / none           8 / 0..7       (default setting for program)
       1 / 0              7 / 1..7
       2 / 0..1           6 / 2..7
       3 / 0..2           5 / 3..7
       4 / 0..3           4 / 4..7            
       5 / 0..4           3 / 5..7
       6 / 0..5           2 / 6..7
       7 / 0..6           1 / 7
       8 / 0..7           0 / none

  Servos are connected to the 3-pin headers located above the Options swtich (4-pos DIP swtich)
  -- high power/high current servos may require external power
  -- contact EFX-TEK for guidance

  LEDs (or other compatible devices) should be connected to the OUTx terminals.
  -- do not use OUTx channels that are assigned to servos

  For playback of exported files the EFX-TEK microSD adapter must be connected to the option
  bins located just above the Options switch.  All Option switches must be moved to the OFF
  postion for microSD card access.

  This program uses the DMX.8 address switch for configuration

    A8 (256)  Off = Live play, On = Play from uSD

}}


con

  PCB_REV = "C"


con

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' 5MHz crystal

  CLK_FREQ = ((_clkmode - xtal1) >> 6) * _xinfreq
  MS_001   = CLK_FREQ / 1_000
  US_001   = CLK_FREQ / 1_000_000
  

con

  RX1     = 31                                                  ' programming / debug port
  TX1     = 30
  
  SDA     = 29                                                  ' eeprom
  SCL     = 28

  OUT_EN  = 27                                                  ' outputs enable pin (low)     
  
  SIO     = 26                                                  ' TTL serial io (for AP-16+)
  
  RX2     = 23                                                  ' RS-485 port
  RX2_EN  = 22
  TX2_EN  = 21
  TX2     = 20

  R_LED   = 18                                                  ' bi-color LED
  G_LED   = 17

  OUT7    = 15                                                  ' outputs
  OUT6    = 14   
  OUT5    = 13 
  OUT4    = 12
  OUT3    = 11
  OUT2    = 10   
  OUT1    =  9 
  OUT0    =  8
          
  OPT_BR  =  7                                                  ' options swithes
  OPT_A1  =  6   
  OPT_A0  =  5
  OPT_SM  =  4

  DMX_A8  =  3                                                  ' high-bit off DMX address

  DO_165  =  2                                                  ' control lines for 74x165
  CLK_165 =  1 
  LD_165  =  0


  ' for EFX-TEK uSD adapter module (#99002)
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR                                       
  SD_DI   =  OPT_A1  
  SD_CLK  =  OPT_A0  
  SD_DO   =  OPT_SM


  PGM_MD  = DMX_A8                                              ' program mode switch


con

   #0, NO, YES

   #0, OFF, RED, GRN, YEL                                       ' for r/g led

   #0, NOCON, RX, TX, RXTX                                      ' for rs-485

   #0, LSBFIRST, MSBFIRST                                       ' for 74x165 shift-register                                               


con

  ' Note that this program is setup for a single HC-8+ so the
  ' maximum channel count is 8.  Servos are on the low channels
  ' so that number should be set first; the number of dimmer
  ' channels is determined automatically.
  
  CHANNELS = 8                                                  ' fixed for single HC-8+
  NSERVOS  = 0                                                  ' servo count, 0 to 8
  NDIMMERS = CHANNELS - NSERVOS                                 ' dimmers are calculated

  SVO_MIN  = 0_600                                              ' for 180-degree servos
  SVO_MAX  = 2_400
  
  MD_LIVE  = 0                                                  ' program modes
  MD_PLAY  = 1

  BAUD     = 115_200                                            ' from Vixen in live mode

  USE_AP16 = 0                                                  ' 1 for audio start in uSD mode
 
  
obj

  sd     : "fsrw"                                               ' SD card IO
  rr     : "realrandom"                                         ' real random from hardware
  prng   : "jm_prng"                                            ' pseudo-random number generator
  servos : "jm_servo8"                                          ' servo driver
  leds   : "jm_pwm8"                                            ' pwm driver for dimmer outputs
  serial : "fullduplexserial64"                                 ' serial input (from Vixen)
  audio  : "fullduplexserial64"                                 ' for serial commands to AP-16+
      

var

  word  chcount                                                 ' channels in Vixen project
  word  eventms                                                 ' event timing; milliseconds
  word  frames                                                  ' frames (events) in project

  byte  framebuf[CHANNELS]                                      ' channel read from uSD card
  

pub main

  setup_io                                                      ' configure HC-8+ io pins

  {
  rr.start                                                      ' load hardware randomizer
  prng.seed(rr.random, rr.random, rr.random, rr.random, rr.random)  
  rr.stop                                                       ' unload hardware randomize
  }

  if (NSERVOS > 0)
    servos.start(NSERVOS, OUT0)                                 ' configure servos (use headers)

  if (NDIMMERS > 0)
    leds.start(NDIMMERS, OUT0+NSERVOS)                          ' configure leds(use OUTx)

  serial.start(RX1, TX1, %0000, BAUD)                           ' start serial input

  if (USE_AP16 == YES)
    audio.start(SIO, SIO, %1100, 38_400)                        ' serial for AP-16+ commands
    pause(2500)                                                 ' give AP-16+ time to boot up 
    
  repeat
    case ina[PGM_MD]                                            ' check mode switch
      MD_LIVE : run_live
      MD_PLAY : play_from_sd
    

pub run_live | sync, level, ch

'' Live mode requires CHANNELS constant to be set
'' -- timeouts used to allow re-test of mode switch

  pause(10)                                                     ' let serial lines settle
  serial.rxflush                                                ' clean-up noise

  repeat
    set_rgled(RED)                                              ' indicate live mode 

    repeat
      level := serial.rxtime(3)                                 ' wait for 3ms idle period
    until (level < 0)

    repeat                                                      ' wait for 1st channel
      if (ina[PGM_MD] == MD_PLAY)                               ' check mode status
        set_rgled(OFF)                                          ' clear live led
        return 0                                                ' back to top 

      level := serial.rxcheck
    until (level => 0)
    update_channel(0, level)

    repeat ch from 1 to CHANNELS-1                              ' rx other channels channels
      level := serial.rxtime(1)                                 ' use timeout to detect stop
      if (level => 0)
        update_channel(ch, level)
      else
        quit
          

pub update_channel(ch, value) 

'' Update channel driver with live or recorded value

  if (ch < NSERVOS)                                             ' servo channel?
    value := scale(value, 255, SVO_MIN, SVO_MAX)                ' convert to servo position 
    servos.set(ch, value, 0)                                    ' update servo
                                                           
  else                                                          ' dimmer channel
    ch -= NSERVOS                                               ' align for dimmer outputs
    leds.set(ch, value)  


dat
                                                                
ShowFile                byte    "show0.bin", 0                  ' show file name


pub play_from_sd | show, t, ch

  repeat
    if (mount_sd)                                               ' if mounted
      set_rgled(GRN)                                            '  read to play
      quit                                                      
    else
      repeat 5                                                  ' red flash before retry
        set_rgled(RED)
        pause(500)
        set_rgled(OFF)
        pause(500)

  repeat
    repeat
      repeat
        if (ina[PGM_MD] == MD_LIVE)                             ' check mode status
          set_rgled(OFF)                                        ' clear play led
          return 0                                              ' back to top
                                                        
        show := ttl_pin                                         ' check for show input
      until (show => 0)

      byte[@ShowFile][4] := "0" + show                          ' set show name     
      if (open_file(@ShowFile, "r"))                            ' file on uSD?
        quit                                                    ' yes, play it
      else
        pause(250)                                              ' allow button release  

    ' send audio start (if enabled)

    if (USE_AP16 == YES)
      audio.str(string("!AP16"))                                ' ap-16+
      audio.tx(%00)                                             ' at address %00
      audio.str(string("PS"))                                   ' play sfx file
      audio.tx(show)                                            ' 0 to 7
      audio.tx(1)                                               ' play once

      ' *** may need to add timing pad here ***      
        
    ' play the show
    
    readbuf(@chcount, 6)                                        ' read file header

    t := cnt                                                    ' sync timer
    repeat frames
      if (readbuf(@framebuf, chcount))                          ' good frame read?
        repeat ch from 0 to chcount-1                           ' copy frame data to dmx
          update_channel(ch, framebuf[ch])                      ' update channel driver  
        waitcnt(t += (MS_001 * eventms))                        ' let frame time expire
        toggle(G_LED)                                           ' flash while playing

      else
        quit

    set_rgled(OFF)                                              ' done

       
pub scale(raw, range, minout, maxout)

'' Scales raw (0 to range) value to new range: minout to maxout

  raw := 0 #> raw <# range                                      ' constrain input

  if (minout < maxout)                                          ' if good range
    return ((raw * (maxout - minout)) / range) + minout
  else
    return raw  


con

  { ================= }
  {                   }
  {  F I L E   I / 0  } 
  {                   }
  { ================= }


var

  long  mounted                                                 ' true when SD mounted


pub mount_sd | check

'' Attempt to mount SD card
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  if (check => 0)
    mounted := true
  else
    mounted := false

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub open_file(spntr, mode) | check

'' Attempt to open file
'' -- spntr is pointer to filename (z-string)
'' -- mode is single character, (e.g., "r" for read)
'' -- sd card must be mounted before open

  if (mounted)
    check := \sd.popen(spntr, mode)                             ' attempt to open
    if (check => 0)
      if ((mode == "r") or (mode == "R"))
        \sd.seek(0)                                             ' force sector read 
      return true
    else
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub readln(spntr, n) | c, len

'' Reads line of text from open text file
'' -- terminated by LF or EOF
'' -- spntr is address of string buffer
'' -- n is maximum number of characters to read from line

  len := 0                                                     ' index into string

  repeat
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else
        quit

    if (c <> 10)                                                ' if not EOL char
      byte[spntr++] := c                                        '  add to buffer
      len++
    else                                                        ' if LF
      quit                                                      '  we're done

  byte[spntr] := 0                                              ' terminate end of line

  return len
  

pub writeln(spntr, usecr, uself)

'' Writes z-string to open text file
'' -- set usecr to true to add CR
'' -- set uself to true to add LF

  \sd.pputs(spntr)
  if (usecr)
    \sd.pputc(13)
  if (uself)
    \sd.pputc(10)


pub readbuf(bpntr, n) | check

'' Reads n bytes from file to buffer at bpntr
'' -- returns true if good read

  check := \sd.pread(bpntr, n)
  if (check == n)
    return true
  else
    return false 


pub writebuf(bpntr, n) | check   

'' Writes n bytes to file from buffer at bpntr
'' -- returns true if good read

  check := \sd.pwrite(bpntr, n)
  if (check == n)
    return true
  else
    return false
     

con

  { ================================= }
  {                                   }
  {  S U P P O R T   R O U T I N E S  } 
  {                                   }
  { ================================= }


var

  long  lastscan                                                ' ttl/DMX input scan


pub setup_io

'' Configure for basic IO control

  outa[OUT7..OUT0] := %0000_0000                                ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                ' output mode                                  
  low(OUT_EN)                                                   ' enable 74x245 

  high(LD_165)                                                  ' setup x165 io pins
  low(CLK_165)
  input(DO_165)
  
  set_rs485(NOCON)                                              ' RS-485 off


pub pause(ms) | t

'' Delay program in milliseconds

  if (ms < 1)                                                   ' delay must be at least 1
    return
  else
    t := cnt - 1792                                             ' sync with system counter
    repeat ms                                                   ' run delay
      waitcnt(t += MS_001)


pub high(pin)

'' Makes pin output and high

  outa[pin] := 1
  dira[pin] := 1


pub low(pin)

'' Makes pin output and low

  outa[pin] := 0
  dira[pin] := 1


pub toggle(pin)

'' Toggles output pin

  !outa[pin]                                                    ' invert state
  dira[pin] := 1  
  

pub input(pin)

  dira[pin] := 0

  return ina[pin]


pub ttl_pin | btns

'' Reads TTL header for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed

  btns := ttl_inputs(true)                                      ' read ttl inputs

  if (bit_count(btns) == 1)                                     ' only one button pressed?
    return bit_pos(btns, LSBFIRST)                              '  yes, return button #
  else
    return -1                                                   '  none or bad input


pub ttl_inputs(rescan) 

'' Return TTL inputs status
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss
   
  return (lastscan & $FF)                                       ' return TTL input bits


pub dmx_address(rescan) 

'' Return DMX address switch setting
'' -- updates global var "lastscan" if rescan is true

  if (rescan)
    lastscan := x165_in(16, MSBFIRST) | (ina[DMX_A8] << 16)     ' read in0..in7, dmx addresss

  return (lastscan >> 8) & $01FF                                ' return DMX address bits
 

pub x165_in(bits, mode) | tmp165

'' Input value from 74x165(s)      
                                                                                          
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                                                         
                                                                                          
  tmp165 := 0                                                   ' clear workspace
  bits := 1 #> bits <# 32                                       ' limit to legal value         
  repeat bits                                                   ' read n bits         
    tmp165 := (tmp165 << 1) | ina[DO_165]                       ' get new bit             
    outa[CLK_165] := 1                                          ' blip clock              
    outa[CLK_165] := 0  

  if (mode == LSBFIRST)                                         ' LSBFIRST result?
    tmp165 ><= bits                                             ' reverse bits
   
  return tmp165
  

pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode

  return !ina[OPT_BR..OPT_SM]                                   ' return state (1 = on)


pub set_rgled(state)

'' Sets R/G LED
'' -- direct control only
'' -- for red/green/yellow use "efx_rgled" driver and methods

  case state
    OFF : outa[R_LED..G_LED] := %00
    RED : outa[R_LED..G_LED] := %10
    GRN : outa[R_LED..G_LED] := %01
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires PASM driver

  dira[R_LED..G_LED] := %11
      

pub set_rs485(state)

'' Sets RS-485 RX and TX enable inputs

  case state
    NOCON : outa[RX2_EN..TX2_EN] := %10
    RX    : outa[RX2_EN..TX2_EN] := %00 
    TX    : outa[RX2_EN..TX2_EN] := %11   
    RXTX  : outa[RX2_EN..TX2_EN] := %01

  dira[RX2_EN..TX2_EN] := %11


pub bit_count(value) | bc

'' Returns # of bits set (1) in value

  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (value & %1)                                          ' add bit value to count
    value >>= 1                                                 ' next bit

  return bc


pub bit_pos(value, mode)

'' Returns position of 1st bit set (0..31)
'' -- mode 0 to scan from lsb, mode 1 to scan from msb 
'' -- -1 = no bits set

  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1

  else
    if (mode == LSBFIRST)                                       ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)
    else
      return (>|value - 1)


pub bit_val(value, pos)

'' Returns bit (0..1) from pos (0..31) in value 

  if ((pos => 0) and (pos =< 31))
    return (value >> pos) & 1
  else
    return 0


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      